#include "conf.h"
#include "igmprt.h"
#include "igmp.h"
#include "udp.h"
#include "csman.h"
#include "csid/csid_local.h"
#include "csid/csid_misc.h"
#include "unilog.h"

/*version and isquerier variable from the config file*/
int version, querier;
//unsigned long upstream;

// for udp_traffic_thread
#define RCV_BUF_SIZE MAX_MSGBUFSIZE
unsigned char *rcv_buf;

// for parse igmpv3 report
unsigned char *report_buf;


void igmp_info_print(igmp_router_t *router)
{	
	igmp_interface_t *ifp;
	igmp_group_t *gp;
	igmp_src_t *src;
	igmp_rep_t *rep;
	printf("\nIGMPv3 Table\n");
	printf("\n%-14s %-9s %-14s %-5s %-14s %-14s\n","Interface","Version","Groups","Mode","Sources","Members");
	printf("____________________________________________________________________________\n");
	for (ifp=router->igmprt_interfaces;ifp;ifp=(igmp_interface_t *)ifp->igmpi_next) {
		printf("%-14s 0x%x\n",inet_ntoa(ifp->igmpi_addr),ifp->igmpi_version);
		if (ifp->igmpi_groups != NULL) {
			for(gp=ifp->igmpi_groups;gp;gp=(igmp_group_t*)gp->igmpg_next) {
				printf("%32s %11s\n",inet_ntoa(gp->igmpg_addr),(gp->igmpg_fmode == IGMP_FMODE_INCLUDE)? "INCL":"EXCL");
				if (gp->igmpg_sources != NULL) {
		    	for (src=gp->igmpg_sources;src;src=(igmp_src_t *)src->igmps_next) {
						printf("%50s\n",inet_ntoa(src->igmps_source));
					}
				}
				if (gp->igmpg_members != NULL) {
					for (rep=gp->igmpg_members;rep;rep=(igmp_rep_t *)rep->igmpr_next) {
						printf("%70s\n",inet_ntoa(rep->igmpr_addr));
					}
				}
				else {
					printf("\n");
				}
			}
		}
		else {
			printf("\n");
		}
	}
	printf("\n");
}

/*
 * igmp_rep_t* igmp_rep_create()
 * Cree un enregistrement d'un membre d'un groupe
 */
igmp_rep_t*
igmp_rep_create(struct in_addr srcaddr)
{
	igmp_rep_t *rep;

	if ((rep = (igmp_rep_t*)malloc(sizeof(*rep)))) {
	  rep->igmpr_addr.s_addr = srcaddr.s_addr;
	  rep->igmpr_next = NULL;
	}
	
	return rep;
}

/*
 * void igmp_rep_cleanup()
 * Effacer un enregistrement d'un membre 
 */
void
igmp_rep_cleanup(igmp_group_t *gp, igmp_rep_t *rep)
{
	igmp_rep_t *re;

	assert(rep != NULL);
	assert(gp  != NULL);
	if (rep != gp->igmpg_members) {
	  for (re=gp->igmpg_members; (igmp_rep_t*)re->igmpr_next!=rep; re=(igmp_rep_t*)re->igmpr_next);		
	  re->igmpr_next = rep->igmpr_next;
	  free(rep);
	}
	else {
	  /*delete the head*/
	  re = gp->igmpg_members;
	  gp->igmpg_members = (igmp_rep_t*)re->igmpr_next;
	  free(re);
	}
	LOG((LOG_DEBUG, "igmp_rep_cleanup: %s\n", inet_ntoa(rep->igmpr_addr)));
}

/*
 * void igmp_rep_print()
 * Imprimer un enregistrement d'un membre
 */
void
igmp_rep_print(
	igmp_rep_t *rep)
{
  printf("\t\tmembre:%-16s\n", 
	 inet_ntoa(rep->igmpr_addr));
}
/*
 * igmp_rep_t* igmp_group_rep_lookup()
 * Lookup a membre in the sourcetable of a group
 */
igmp_rep_t*
igmp_group_rep_lookup(
	igmp_group_t *gp,
	struct in_addr srcaddr)
{
  igmp_rep_t *rep;
  
  assert(gp != NULL);
  for (rep=(igmp_rep_t*)gp->igmpg_members; rep; rep=(igmp_rep_t*)rep->igmpr_next) {
    if (rep->igmpr_addr.s_addr == srcaddr.s_addr)
      return rep;
  }
  
  return NULL; 
}

/*
 * igmp_rep_t* igmp_group_rep_add()
 * Add a member to the set of sources of a group
 */
igmp_rep_t*
igmp_group_rep_add(
	igmp_group_t *gp,
	struct in_addr srcaddr)
{
	igmp_rep_t *rep;

	assert(gp != NULL);
	/* Return the source if it's already present */
	if (rep = igmp_group_rep_lookup(gp, srcaddr)) {
	  return rep;
	}
	/* Create the source and add to the set */ 
	if (rep = igmp_rep_create(srcaddr)) {
	  rep->igmpr_next = (igmp_rep_t*)gp->igmpg_members;
	  gp->igmpg_members = rep;
	}
	
	return rep;
}
/******************************************************************************
 * igmp source routines 
 *****************************************************************************/
/*
 * igm_src_t* igmp_src_create()
 * Cree un enregistrement d'une source
 */
igmp_src_t*
igmp_src_create(
	struct in_addr srcaddr)
{
	igmp_src_t* src;

	if (src = (igmp_src_t*)malloc(sizeof(*src))) {
	  src->igmps_source.s_addr = srcaddr.s_addr;
	  src->igmps_timer         = 0;
	  src->igmps_fstate        = TRUE;
	  src->igmps_next          = NULL;
	}
	
	return src;
}

/*
 * void igmp_src_cleanup()
 * Effacer un enregistrement d'une source
 */
void
igmp_src_cleanup(igmp_group_t *gp, igmp_src_t *src)
{
	igmp_src_t *sr;
	
	assert(src != NULL);
	assert (gp != NULL);
	
	if (src != gp->igmpg_sources) {
	  for (sr=gp->igmpg_sources; (igmp_src_t*)sr->igmps_next!=src; sr=(igmp_src_t*)sr->igmps_next);
	  sr->igmps_next = src->igmps_next;
	  free(src);
	}
	else {
	  /*delete the head*/
	  gp->igmpg_sources = (igmp_src_t*)src->igmps_next;
	  free(src);
	}
	LOG((LOG_DEBUG, "igmp_src_cleanup: %s\n", inet_ntoa(src->igmps_source)));
}

/*
 * void igmp_src_print()
 * Imprimer un enregistrement d'une source
 */
void
igmp_src_print(
	igmp_src_t *src)
{
  printf("\t\tsource:%-16s  %d\n", 
	 inet_ntoa(src->igmps_source),src->igmps_timer);
}
/*
 * igmp_src_t* igmp_group_src_lookup()
 * Lookup a source in the sourcetable of a group
 */
igmp_src_t*
igmp_group_src_lookup(
	igmp_group_t *gp,
	struct in_addr srcaddr)
{
  igmp_src_t *src;
  
	if (gp != NULL) {
		for (src=gp->igmpg_sources; src; src=(igmp_src_t*)src->igmps_next) {
			if (src->igmps_source.s_addr == srcaddr.s_addr)
				return src;
		}
	}
  return NULL; 
}

/*
 * igmp_src_t* igmp_group_src_add()
 * Add a source to the set of sources of a group
 */
igmp_src_t*
igmp_group_src_add(igmp_group_t *gp, struct in_addr srcaddr)
{
	igmp_src_t* src;

	assert(gp != NULL);
	
	/* Return the source if it's already present */
	if (src = igmp_group_src_lookup(gp, srcaddr))
  	return src;
	  	
	/* Create the source and add to the set */ 
	if (src = igmp_src_create(srcaddr)) {
	  src->igmps_next = (struct igmp_src_t*)gp->igmpg_sources;
	  gp->igmpg_sources = src;
	}
	
	return src;
}



/******************************************************************************
 * igmp group routines 
 *****************************************************************************/
/*
 * igmp_group_t* igmp_group_create()
 * Create a group record
 */
igmp_group_t*
igmp_group_create(struct in_addr groupaddr)
{
	igmp_group_t* gp;
	
	gp = NULL;
	if (gp = (igmp_group_t*)malloc(sizeof(*gp))) {
		gp->igmpg_addr.s_addr = groupaddr.s_addr;
		gp->igmpg_fmode 			= IGMP_FMODE_INCLUDE;
		gp->igmpg_version			= IGMP_VERSION_3;
		gp->igmpg_timer 			= 0;
		gp->igmpg_sources 		= NULL;
		gp->igmpg_members 		= NULL;
		gp->igmpg_next 				= NULL;
	}
	
	return gp;
}

/*
 * void igmp_group_cleanup()
 * Cleanup a group record
 */
void
igmp_group_cleanup(igmp_interface_t *ifp, igmp_group_t* gp)
{
	igmp_group_t *g;
	
	assert(gp != NULL);
	assert(ifp != NULL);
	
	if (ifp->igmpi_groups != gp) {
	  g=ifp->igmpi_groups;
	  while ((igmp_group_t*)g->igmpg_next != gp) {		
			g=(igmp_group_t*)g->igmpg_next;
		}
	  g->igmpg_next=gp->igmpg_next;		
	  free(gp);
	}
	else {/*delete the head*/
	  g=ifp->igmpi_groups;
	  ifp->igmpi_groups = (igmp_group_t*)g->igmpg_next;
	  free(g);
	} 
	LOG((LOG_DEBUG, "igmp_group_cleanup: %s\n", inet_ntoa(gp->igmpg_addr)));
}

/*
 * void igmp_group_print()
 * Print a group record
 */
void
igmp_group_print(igmp_group_t* gp)
{
  igmp_src_t *src;
  igmp_rep_t *rep;
  
  printf("  %-16s %s %d\n", inet_ntoa(gp->igmpg_addr),
	 			(gp->igmpg_fmode==IGMP_FMODE_EXCLUDE)?"exclude":"include", gp->igmpg_timer);
  if (gp->igmpg_sources != NULL) {
    for (src=gp->igmpg_sources; src; src=(igmp_src_t*)src->igmps_next)
      printf("source : %s timer : %d\n",inet_ntoa(src->igmps_source.s_addr),src->igmps_timer);
  }
  if (gp->igmpg_members != NULL) {
    for (rep=gp->igmpg_members; rep; rep=(igmp_rep_t*)rep->igmpr_next)
      printf("member : %s \n",inet_ntoa(rep->igmpr_addr.s_addr));
  }
}

/******************************************************************************
 * igmp interface routines 
 *****************************************************************************/
/*
 * igmp_interface_t* igmp_interface_create()
 * Create and initialize interface record
 */
igmp_interface_t*
igmp_interface_create(struct in_addr ifaddr, char *ifname, vifi_t index)
{
  igmp_interface_t* ifp;
  struct ip_mreq mreq;
  char ra[4];
  int i, prom;
  short flags;	  
	  
  if (((ifp = (igmp_interface_t*) malloc(sizeof(*ifp)))) == NULL)
	  return NULL;
	
  /* Allocate a buffer to receive igmp messages */ 		
  ifp->igmpi_bufsize = get_interface_mtu(ifname);
  if (ifp->igmpi_bufsize < 0)
	  ifp->igmpi_bufsize = MAX_MSGBUFSIZE;

  /* XXX Should make buffer slightly bigger for ip header */
  if ((ifp->igmpi_buf = (char*) malloc(ifp->igmpi_bufsize)) == NULL) {
	  free(ifp);
	  return NULL;
  }	
	
  /* Create a raw igmp socket */
  ifp->igmpi_socket = socket(AF_INET, SOCK_RAW, IPPROTO_IGMP);
  if (ifp->igmpi_socket < 0) {
    printf("can't create socket \n");
    free(ifp->igmpi_buf);
    free(ifp);
    return NULL;
  }
		
  /* Initialize all fields */
  ifp->igmpi_addr.s_addr = ifaddr.s_addr;
  if (ifp->igmpi_addr.s_addr == upstream) {
	  ifp->igmpi_type = UPSTREAM;
	}
  else {
	  ifp->igmpi_type = DOWNSTREAM;
	}
  strncpy(ifp->igmpi_name, ifname, IFNAMSIZ);
  ifp->igmpi_groups    = NULL;
  ifp->sch_group_query = NULL;		
  ifp->igmpi_isquerier = querier;
  ifp->igmpi_version   = version;
  ifp->igmpi_qi        = IGMP_DEF_QI;
  ifp->igmpi_qri       = IGMP_DEF_QRI;
  ifp->igmpi_rv        = IGMP_DEF_RV;
  ifp->igmpi_gmi       = IGMP_GMI;
  ifp->igmpi_ti_qi     = 0;
  ifp->igmpi_next      = NULL;

	/* Set router alert */
	ra[0] = 148;
	ra[1] = 4;
	ra[2] = 0;
	ra[3] = 0;
	setsockopt(ifp->igmpi_socket, IPPROTO_IP, IP_OPTIONS, ra, 4);

	/* Set reuseaddr, ttl, loopback and set outgoing interface */
	i = 1;
	setsockopt(ifp->igmpi_socket, SOL_SOCKET, SO_REUSEADDR, (void*)&i, sizeof(i));
	i = 1;
	setsockopt(ifp->igmpi_socket, IPPROTO_IP, IP_MULTICAST_TTL, (void*)&i, sizeof(i));	
	i = 0;
	setsockopt(ifp->igmpi_socket, IPPROTO_IP, IP_MULTICAST_LOOP, (void*)&i, sizeof(i));

	setsockopt(ifp->igmpi_socket, IPPROTO_IP, IP_MULTICAST_IF, (void*)&ifaddr, sizeof(ifaddr));

#if 0
	/* Add membership to ALL_ROUTERS and ALL_ROUTERS_V3 on this interface */
	mreq.imr_multiaddr.s_addr = inet_addr(IGMP_ALL_ROUTERS);
	mreq.imr_interface.s_addr = ifaddr.s_addr;/*htonl(0);*/
	setsockopt(ifp->igmpi_socket, IPPROTO_IP, IP_ADD_MEMBERSHIP, (void*)&mreq, sizeof(mreq));
	
	mreq.imr_multiaddr.s_addr = inet_addr(IGMP_ALL_ROUTERS_V3);
	mreq.imr_interface.s_addr = ifaddr.s_addr;/*htonl(0);*/
	setsockopt(ifp->igmpi_socket, IPPROTO_IP, IP_ADD_MEMBERSHIP, (void*)&mreq, sizeof(mreq));
#endif
	
	/* Tell the kernel this interface belongs to a multicast router */
	mrouter_onoff(ifp->igmpi_socket,1);
	ifp->igmpi_index = index;
	if (ifp->igmpi_addr.s_addr == upstream)
	  upstreamix = ifp->igmpi_index;
	
	/* Set the interface flags to receive all multicast packets */
	ifp->igmpi_save_flags = get_interface_flags(ifname);
	if (ifp->igmpi_save_flags != -1) {
	  set_interface_flags(ifname, ifp->igmpi_save_flags | IFF_ALLMULTI | IFF_PROMISC);
	}

	return ifp;
}

/*
 * void igmp_interface_cleanup()
 * Cleanup an interface 
 */
void
igmp_interface_cleanup(igmp_interface_t* ifp)
{
	igmp_group_t* gp,*gp2;

	assert(ifp != NULL);
	LOG((LOG_DEBUG, "igmp_interface_cleanup: %s\n", inet_ntoa(ifp->igmpi_addr)));

	/* Cleanup all groups */
	gp=ifp->igmpi_groups;
	ifp->igmpi_groups = NULL;
	while(gp != NULL) {
		gp2=gp;
		gp=gp->igmpg_next;
		free(gp2);
	}
	/* Tell the kernel the multicast router is no more */
	mrouter_onoff(ifp->igmpi_socket, 0);
	close(ifp->igmpi_socket);
 	free(ifp->igmpi_buf);

	/* If we managed to get/set the interface flags, revert */
	if (ifp->igmpi_save_flags != -1)
		set_interface_flags(ifp->igmpi_name, ifp->igmpi_save_flags);
			
	free(ifp);
}

/*
 * igmp_group_t* igmp_interface_group_add()
 * Add a group to the set of groups of an interface
 */
igmp_group_t*
igmp_interface_group_add(
	igmp_router_t* router,
	igmp_interface_t *ifp,
	struct in_addr groupaddr)
{
	igmp_interface_t *upstream_interface;
	igmp_group_t *gp;
	struct ip_mreq mreq;	

	assert(ifp != NULL);
		
	/* Return this group if it already exists. */
	if (gp = igmp_interface_group_lookup(ifp, groupaddr))
		return gp;
	
	/* Create group and add to table */
	if (gp = igmp_group_create(groupaddr)) {
		gp->igmpg_next = ifp->igmpi_groups;
		ifp->igmpi_groups = gp;

		/* Add group to the set of groups of upstream interface */
		mreq.imr_multiaddr.s_addr = groupaddr.s_addr;
		mreq.imr_interface.s_addr = upstream;
		if ((VALID_ADDR(mreq.imr_multiaddr)))
			if (setsockopt(router->igmprt_up_socket, IPPROTO_IP, IP_ADD_MEMBERSHIP, (void *)&mreq, sizeof(mreq)) < 0)
	    	perror("setsockopt - IP_ADD_MEMBERSHIP");
	}
	
	return gp;
}

/*
 * igmp_group_t* igmp_interface_group_lookup()
 * Lookup a group in the grouptable of an interface
 */
igmp_group_t*
igmp_interface_group_lookup(igmp_interface_t *ifp, struct in_addr groupaddr)
{
	igmp_group_t* gp;

	assert(ifp != NULL);
	
	for (gp = ifp->igmpi_groups; gp; gp = gp->igmpg_next)
		if (gp->igmpg_addr.s_addr == groupaddr.s_addr)
			return gp;

	return NULL; 
}

/*
 * void igmp_interface_membership_report_v12()
 * Process a v1/v2 membership report
 */
void
igmp_interface_membership_report_v12(
	igmp_router_t* router,	
	igmp_interface_t* ifp,
	struct in_addr src,
	igmpr_t* report,
	int len)
{
	igmp_group_t* gp;
	igmp_rep_t *rep;
	igmp_src_t *source;

	/* Ignore a report for a non-multicast address */ 
	if (!IN_MULTICAST(ntohl(report->igmpr_group.s_addr)))
		return;
		
	/* Find the group and add it if not present */
	if ((gp = igmp_interface_group_add(router,ifp,report->igmpr_group)) == NULL)
		return;
  
  /* find the member and add it if not present */
	rep = igmp_group_rep_add(gp, src);
	
	/* Consider this to be a v3 is_ex{} report */
	igmp_group_handle_isex(router, ifp, gp, 0, NULL);
}

/*
 * void igmp_interface_membership_leave_v2()
 * Process a v2 membership leave
 */
void
igmp_interface_membership_leave_v2(
	igmp_router_t* router,	
	igmp_interface_t* ifp,
	struct in_addr src,
	igmpr_t* report,
	int len)
{
	igmp_group_t *gp;
	igmp_rep_t *walker;
	int fd = open_csman(NULL, 0), quickleave=0, mcount=0;

	/* Ignore a report for a non-multicast address */ 
	if (!IN_MULTICAST(ntohl(report->igmpr_group.s_addr)))
		return;
		
	/* Find the group */
	if (!(gp = igmp_interface_group_lookup(ifp, report->igmpr_group)))
		return;
  
	read_csman(fd, CSID_C_MISC_IGMP_QUICKLEAVE, &quickleave, sizeof(quickleave), CSM_R_ZERO);
	close_csman(fd);
	if (quickleave) {
		for (walker=gp->igmpg_members; walker; walker=walker->igmpr_next)
			mcount++;
		
		if (mcount <= 1) {
			gp->igmpg_timer = 0;
			return;
		}
	}
	gp->igmpg_timer = IGMP_DEF_LI;
	igmp_group_handle_toin(router, ifp, gp, 0, NULL);
	
	return;
}

/*
 * void igmp_interface_membership_report_v3()
 * Process a v3 membership report
 */
void
igmp_interface_membership_report_v3(
  igmp_router_t* router,
  igmp_interface_t* ifp,
  struct in_addr src,
  void* pack,
  int len)
{
	igmp_report_t *report;
	igmp_group_t *gp;
  igmp_rep_t *rep;
	u_short numsrc;
	u_short numgrps, numsrcs;
	u_char type;
	int i;

  if ((ifp->igmpi_addr.s_addr == upstream) && (src.s_addr != upstream))
   	forward_upstream = 1;

	// Parse igmpv3 report packet
	report = igmp_parse_pack(pack);
	// End parse igmpv3 report

	numgrps = ntohs(report->igmpr_numgrps);
	for (i=0; i<numgrps; ++i) {
		if (!IN_MULTICAST(ntohl(report->igmpr_group[i].igmpg_group.s_addr)))
  	 	return;
  	 	
		if ((gp = igmp_interface_group_add(router, ifp, report->igmpr_group[i].igmpg_group)) == NULL)
  	 	return;

		//if (src.s_addr != ifp->igmpi_addr.s_addr)
		rep = igmp_group_rep_add(gp, src);
		
		/* Find the group record type */
		type   = (u_char)report->igmpr_group[i].igmpg_type;
		numsrc = ntohs(report->igmpr_group[i].igmpg_numsrc);
			
		switch(type) {
			case 1:
	       igmp_group_handle_isin(router,ifp,gp,numsrc,(struct in_addr*)&report->igmpr_group[i].igmpg_sources);
	       break;
	       
			case 2:
	       igmp_group_handle_isex(router,ifp,gp,numsrc,(struct in_addr*)&report->igmpr_group[i].igmpg_sources);
	       break;
	       
			case 3:
	       igmp_group_handle_toin(router,ifp,gp,numsrc,(struct in_addr*)&report->igmpr_group[i].igmpg_sources);
		     break;
		     
			case 4:
	       igmp_group_handle_toex(router,ifp,gp,numsrc,(struct in_addr*)&report->igmpr_group[i].igmpg_sources);
	       break;
	       
			case 5:
	       igmp_group_handle_allow(router,ifp,gp,numsrc,(struct in_addr*)&report->igmpr_group[i].igmpg_sources);
	       break;
	       
			case 6:
	       igmp_group_handle_block(router,ifp,gp,numsrc,(struct in_addr*)&report->igmpr_group[i].igmpg_sources);
	       break;
	       
			default:
	       LOG((LOG_INFO, "igmp_interface_membership_report_v3: group record type undefined"));
	       break;
		}
	}
}

/*
 * void igmp_interface_print()
 * Print status of an interface 
 */
void
igmp_interface_print(igmp_interface_t* ifp)
{
	igmp_group_t* gp;

	printf(" interface %s, %s ver=0x%x\n",
		inet_ntoa(ifp->igmpi_addr),(ifp->igmpi_type == UPSTREAM)?"UPSTREAM":"DOWNSTREAM", ifp->igmpi_version);
	for (gp = ifp->igmpi_groups; gp; gp = gp->igmpg_next)
		igmp_group_print(gp);
}

/******************************************************************************
 * igmp router routines
 *****************************************************************************/
/*
 * int igmprt_init()
 * Initialize igmp router
 */
int
igmprt_init(igmp_router_t* igmprt)
{
  int i;
		
  igmprt->igmprt_interfaces    = NULL;
  igmprt->igmprt_thr_timer     = 0;
  igmprt->igmprt_thr_input     = 0;
  igmprt->igmprt_thr_udpproxy  = 0;
  igmprt->igmprt_flag_timer    = 0;
  igmprt->igmprt_flag_input    = 0;
  igmprt->igmprt_flag_udpproxy = 0;
		
  /*create datagram socket to tell igmpv3 host about reports to send on upstream interface*/
  //if ((igmprt->igmprt_up_socket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
  if ((igmprt->igmprt_up_socket = socket(AF_INET, SOCK_RAW, IPPROTO_UDP)) < 0)
	  perror("can't open upstream socket");
		
  /*create raw socket to update mfc and vif table*/
  if ((igmprt->igmprt_socket = socket(AF_INET, SOCK_RAW, IPPROTO_IGMP)) < 0)
	  perror("can't open igmp socket");

	i = 1;
	setsockopt(igmprt->igmprt_socket, IPPROTO_IP, IP_MULTICAST_TTL, (void*)&i, sizeof(i));
	i = 0;
	setsockopt(igmprt->igmprt_socket, IPPROTO_IP, IP_MULTICAST_LOOP, (void*)&i, sizeof(i));
 
  forward_upstream = 0;
		
  // malloc for udp_traffic_thread() to receive and forward udp stream.
  rcv_buf = (unsigned char*) malloc(RCV_BUF_SIZE);  
  
  // malloc for igmp_parse_pack() to parse received igmpv3 report
  report_buf = (unsigned char*) malloc(MAX_MSGBUFSIZE);		
  
  return 1;
}

/*
 * void igmprt_destroy()
 * Cleanup the router 
 */
void
igmprt_destroy(igmp_router_t* igmprt)
{
  igmp_interface_t *ifp, *ifp2;
  igmp_group_t *gp;
  igmp_src_t *sp;
  
  // Free malloc spaces
  free(rcv_buf);
  free(report_buf);
  // End
	
  for (ifp=igmprt->igmprt_interfaces; ifp;) {
	  ifp2 = ifp;
	  ifp = ifp->igmpi_next;
	  igmp_interface_cleanup(ifp2);
  }
  LOG((LOG_DETAIL, "destroy igmp router ...\n"));
  igmprt_stop(igmprt);
}

/*
 * interface_t* igmprt_interface_lookup()
 * Lookup a group, identified by the interface address
 */
igmp_interface_t*
igmprt_interface_lookup(igmp_router_t* igmprt, struct in_addr ifaddr)
{
	igmp_interface_t *ifp;

	for (ifp=igmprt->igmprt_interfaces; ifp; ifp=ifp->igmpi_next)
		if (ifp->igmpi_addr.s_addr == ifaddr.s_addr)
			return ifp;
			
	return NULL;
}

/*
 * interface_t* igmprt_interface_lookup_index()
 * Lookup a group, identified by the interface index
 */
igmp_interface_t*
igmprt_interface_lookup_index(igmp_router_t* igmprt, int ifp_index)
{
	igmp_interface_t *ifp;

	for (ifp=igmprt->igmprt_interfaces; ifp; ifp=ifp->igmpi_next)
		if (ifp->igmpi_index == ifp_index)
			return ifp;
	
	return NULL;
}

/*
 * igmp_entry_t* igmprt_group_lookup() 
 * Lookup a group, identified by the interface and group address
 */
igmp_group_t*
igmprt_group_lookup(
	igmp_router_t* igmprt, 
	struct in_addr ifaddr, 
	struct in_addr groupaddr)
{
	igmp_interface_t *ifp; 

	if (ifp = igmprt_interface_lookup(igmprt, ifaddr))
		return igmp_interface_group_lookup(ifp, groupaddr);

	return NULL;
}

/*
 * struct igmp_interface_t* igmprt_interface_add()
 * Add an interface to the interfacetable
 */
igmp_interface_t*
igmprt_interface_add(
	igmp_router_t* igmprt, 
	struct in_addr ifaddr,
	char *ifname,
	vifi_t index)
{
	igmp_interface_t *ifp;

	/* Return the interface if it's already in the set */
	if (ifp = igmprt_interface_lookup(igmprt, ifaddr))
		return ifp;

	/* Create the interface and add to the set */
	if (ifp = igmp_interface_create(ifaddr, ifname,index)) {
		ifp->igmpi_next = igmprt->igmprt_interfaces;
		igmprt->igmprt_interfaces = ifp;
	}
	
	return ifp;
}

/*
 * igmp_group_t* igmprt_group_add()
 * Add a group to the grouptable of some interface
 */
igmp_group_t*
igmprt_group_add(
	igmp_router_t* igmprt, 
	struct in_addr ifaddr, 
	struct in_addr groupaddr)
{
	igmp_interface_t *ifp; 
	igmp_group_t *gp; 

	if (ifp = igmprt_interface_lookup(igmprt, ifaddr))
		return NULL;

	return igmp_interface_group_add(igmprt,ifp, groupaddr);
}

/*
 * void igmprt_timergroup(igmp_router_t* igmprt)
 * Decrement timers and handle whatever has to be done when one expires
 */
void
igmprt_timer(igmp_router_t* igmprt)
{
  igmp_interface_t* ifp;
  igmp_group_t* gp;
  struct in_addr zero;
  
  zero.s_addr = 0;
  
  /* Handle every interface */
  for (ifp = igmprt->igmprt_interfaces; ifp; ifp = ifp->igmpi_next) {
	  /* If we're the querier for this network, handle all querier 
	   * duties */
	  if (ifp->igmpi_isquerier == TRUE) {
	    /* Deal with the general query */
	    if (--(ifp->igmpi_ti_qi) <= 0) {
				ifp->igmpi_ti_qi = ifp->igmpi_qi;
				igmprt_membership_query(igmprt, ifp, &zero, NULL, 0, 0);
	    }
	  }
	  else{
	  	/* If not the querier, deal with other-querier-present timer*/
	   	igmprt_timer_querier(ifp);	
	  }
	  /*handle group timer*/
	  /*LOG((LOG_DEBUG, "handle timer group\n"));*/
	  igmprt_timer_group(igmprt,ifp);
	  /*handle source timer*/
	  /*LOG((LOG_DEBUG, "handle timer source\n"));*/
	  igmprt_timer_source(igmprt,ifp);
	  /*handle scheduled query*/
	  send_sh_query(igmprt,ifp);   	
  }
}

/*
 * void* igmprt_timer_thread(void* arg)
 * The thread that deals with the timer 
 */
void*
igmprt_timer_thread(void* arg)
{
  igmp_router_t* igmprt = (igmp_router_t*) arg;
  
  while (igmprt->igmprt_flag_timer) {
    igmprt_timer(igmprt);
    /* Should be changed to take care of drifting */
    sleep(1);
  }
  pthread_exit(NULL);
  return NULL;
}

/*
 * void igmprt_input()
 * Process an incoming igmp message
 */
void
igmprt_input(igmp_router_t* igmprt, igmp_interface_t* ifp)
{
	igmp_report_t *report;
	igmpv3q_t *query;
	struct ip *iph;
	unsigned char *igmphdr;
	int rcvlen, igmplen, iphlen, dummy = 0;
	int if_index, srsp;

	memset((char *)ifp->igmpi_buf, 0, ifp->igmpi_bufsize);
	rcvlen = recvfrom(igmprt->igmprt_socket, ifp->igmpi_buf, ifp->igmpi_bufsize, 0, NULL, &dummy);
					
	if (rcvlen <= sizeof(struct ip)) {
			return;
	}

	/* Set pointer to start of report */
	iph    = (struct ip *)ifp->igmpi_buf;
	iphlen = iph->ip_hl << 2;

	if (!IN_MULTICAST(ntohl(iph->ip_dst.s_addr)))
  	return;

	if (iph->ip_src.s_addr == downstream)
		return;

	if ((igmplen = rcvlen - iphlen) < IGMP_MINLEN)
		return;

	
	/* Handle the message */
 	igmphdr = ifp->igmpi_buf + iphlen;
	switch (*igmphdr) {
		case IGMP_MEMBERSHIP_QUERY:
			query = (igmpv3q_t*)igmphdr;
			srsp  = IGMP_SRSP(query);
			if (query->igmpq_code == 0) { //v1 query
				LOG((LOG_DEBUG, "igmpv1 query \n"));
				receive_membership_query(ifp, query->igmpq_group, NULL, iph->ip_src.s_addr, 0, srsp);
			}
			else if (igmplen == 8) { //v2 query		
				LOG((LOG_DEBUG, "igmpv2 query \n"));
				receive_membership_query(ifp, query->igmpq_group, NULL, iph->ip_src.s_addr, 0, srsp);
			}
			else if (igmplen >= 12) { //v3 query
				receive_membership_query(ifp, query->igmpq_group, query->igmpq_sources, 
																 iph->ip_src.s_addr, query->igmpq_numsrc, srsp);
			}
			break;
				
		case IGMP_V1_MEMBERSHIP_REPORT:
			// To judge report is coming from LAN side or not.
			if ((iph->ip_src.s_addr & mask) != subnet) {
				return;
			}
			ifp->igmpi_version = IGMP_VERSION_1;
			igmp_interface_membership_report_v12(igmprt, ifp, iph->ip_src, (igmpr_t*)igmphdr, igmplen);
			break;

		case IGMP_V2_MEMBERSHIP_REPORT:
			// To judge report is coming from LAN side or not.
			if ((iph->ip_src.s_addr & mask) != subnet) {
				return;
			}
			ifp->igmpi_version = IGMP_VERSION_2;
			igmp_interface_membership_report_v12(igmprt, ifp, iph->ip_src, (igmpr_t*)igmphdr, igmplen);
			break;
			
		case IGMP_V2_MEMBERSHIP_LEAVE:
			if ((iph->ip_src.s_addr & mask) != subnet) {
				return;			
			}
			ifp->igmpi_version = IGMP_VERSION_2;
			igmp_interface_membership_leave_v2(igmprt, ifp, iph->ip_src, (igmpr_t*)igmphdr, igmplen);
			break;
	
		case IGMP_V3_MEMBERSHIP_REPORT:
			// To judge report is coming from LAN side or not.
			if ((iph->ip_src.s_addr & mask) != subnet) {
				return;			
			}
			ifp->igmpi_version = IGMP_VERSION_3;
  		igmp_interface_membership_report_v3(igmprt, ifp, iph->ip_src, (void*)igmphdr, igmplen);
      break;
	
		default:
			break;
	}
}

/*
 * void* igmprt_input_thread(void* arg)
 * Wait on all interfaces for packets to arrive
 */
void*
igmprt_input_thread(void *arg)
{
	igmp_router_t* igmprt = (igmp_router_t*)arg;
	igmp_interface_t* ifp;
	fd_set allset, rset;
	int n, maxfd;
	
	/* Add the sockets from all interfaces to the set */
	FD_ZERO(&allset);
	maxfd = 0;
	for (ifp=igmprt->igmprt_interfaces; ifp; ifp=ifp->igmpi_next) {
		FD_SET(ifp->igmpi_socket, &allset);
		if (maxfd < ifp->igmpi_socket) {
			maxfd = ifp->igmpi_socket;
		}
	}
	
	if (maxfd == 0) {
		LOG((LOG_INFO, "error: no interfaces available to wait for input\n"));
		return NULL;
	}
		
	/* Wait for data on all sockets */
	while (igmprt->igmprt_flag_input) {
		FD_COPY(&allset, &rset);
		n = select(maxfd+1, &rset, NULL, NULL, NULL);				
		if (n == 0) {
			sleep(1);
		}				
		for (ifp=igmprt->igmprt_interfaces; ifp; ifp=ifp->igmpi_next){
			if (FD_ISSET(ifp->igmpi_socket, &rset)) {
				igmprt_input(igmprt, ifp);
			}	
			if ((--n) == 0) {
				break;
			}
		}
	}
		
	return NULL;
}

/*
 * void igmprt_start(igmp_router_t* igmprt)
 * Start the threads of this router 
 */
void
igmprt_start(igmp_router_t *igmprt)
{
	int err;

	/* Return if already running */
	if (igmprt->igmprt_running)
	  return;

	/* Create and start the timer handling (thread) */
	igmprt->igmprt_flag_timer = 1;
	igmprt->igmprt_thr_timer  = 0;
	if ((err = pthread_create(&igmprt->igmprt_thr_timer,NULL,igmprt_timer_thread,(void*)igmprt)) != 0) {
	  printf("couldn't start timer thread\n");
	}
	
	/* Create and start input handling (thread) */
	igmprt->igmprt_flag_input = 1;
	igmprt->igmprt_thr_input  = 0;
	if ((err = pthread_create(&igmprt->igmprt_thr_input,NULL,igmprt_input_thread,(void*)igmprt)) != 0) {
	  printf("couldn't start input thread\n");
	}
		
	/* Create and start udp traffic handling (thread) */
	igmprt->igmprt_flag_udpproxy = 1;
	igmprt->igmprt_thr_udpproxy  = 0;
	if ((err = pthread_create(&igmprt->igmprt_thr_udpproxy,NULL,udp_traffic_thread,(void*)igmprt)) != 0) {
	  printf("couldn't start udp_traffic thread\n");
	}

	igmprt->igmprt_running = 1;
}

/*
 * void igmprt_stop(igmp_router_t* igmprt)
 * Stop the threads of this router 
 */
void
igmprt_stop(igmp_router_t *igmprt)
{
	igmp_interface_t* ifp;
	int i = 0;

	/* Return if not running */ 
	if (!igmprt->igmprt_running)
		return;

	k_stop_proxy(igmprt->igmprt_socket);

	/* Signal threads to stop */
	igmprt->igmprt_flag_timer    = 0;
	igmprt->igmprt_flag_input    = 0;
	igmprt->igmprt_flag_udpproxy = 0;	

	/* Wait for the threads to finish */
	/* XXX Should use attach */ 
	igmprt->igmprt_running = 0;
	/* Make sure select of  input-thread wakes up */	
	/*if ((ifp = igmprt->igmprt_interfaces))
		write(ifp->igmpi_socket, &i, sizeof(i));*/
	sleep(3);
}

/*
 * void igmprt_print()
 * Print the status of the igmpv3 proxy/router
 */
void
igmprt_print(igmp_router_t *igmprt)
{
	igmp_interface_t* ifp;

	assert(igmprt != NULL);
	printf("igmp router:\n");
	for (ifp=igmprt->igmprt_interfaces; ifp; ifp=ifp->igmpi_next)
		igmp_interface_print(ifp);
}

/******************************************************************************
 * The application
 *****************************************************************************/
#define BUF_CMD	100

igmp_router_t router;
int go_on = 1;
char *pidfile = DEFAULT_PID_FILE_NAME;

void
write_pid(void)
{
	FILE *fp = fopen(pidfile, "w");
	if (fp) {
		fprintf(fp, "%d\n", getpid());
		fclose(fp);
	}
}

void
done(int sig)
{
	igmprt_destroy(&router);
	unlink(pidfile);
	exit(0);
}

void print_usage()
{
	printf("g Multicast Address:show a specific-group state\n");
	printf("s Unicast Address  :show a specific-source state\n");
	printf("a		   :show IGMP Table\n");
	printf("q		   :quit\n");
}

/* Process a command */
int process_cmd (char *cmd)
{
	char *line;
	char group[20], i[20], source[20], snoop[2];
	igmp_group_t *gp;
	igmp_src_t *src;
	struct in_addr interface, gr, srcaddr;
	igmp_interface_t *ifp;
	
	/* skip white spaces */
	line = cmd + 1;
	while (isblank(*line)) ++line;
	switch (*cmd) {
		case '?':
				print_usage();
				return 1;
				
		case 'q':
				return 0;
				
		case 'g':
	  		/* print a group details*/
				if (sscanf(line,"%s",group) != 1){
					printf("\n");
					return 1;
	   		}
	  		/* interface.s_addr=inet_addr(i);*/
	   		gr.s_addr=inet_addr(group);
	   		for (ifp=(&router)->igmprt_interfaces;ifp;ifp=ifp->igmpi_next){
	 				gp=igmprt_group_lookup(&router,ifp->igmpi_addr,gr);
					if (gp != NULL){ 
						printf("interface: %s\n",inet_ntoa(ifp->igmpi_addr));	
	     			igmp_group_print(gp);
					} else{
	 					printf("unknown group\n");
	 				}
	   		}
	      return 1;
	      
		case 's':
	    	/* print a source of a group details*/
	    	if (sscanf(line,"%s",source) !=1){	
					printf("\n");
					return 1;
	    	}
	    	gr.s_addr=inet_addr(group);
		    srcaddr.s_addr=inet_addr(source);	
		    for (ifp=(&router)->igmprt_interfaces;ifp;ifp=ifp->igmpi_next){
					for (gp=ifp->igmpi_groups;gp;gp=gp->igmpg_next){				
						if ((src=igmp_group_src_lookup(gp,srcaddr)) != NULL){
							printf("interface : %s\n",inet_ntoa(ifp->igmpi_addr));
							printf("\tgroup : %s\n",inet_ntoa(gp->igmpg_addr));
							igmp_src_print(src);
						}
					}
				}
	    	return 1;
	    	
		case 'a':
				/*print all details of the router*/
	 			/*igmprt_print(&router);   with this more details*/
	 			igmp_info_print(&router);  /* but this more beautiful*/
				return 1;
							
		default:
	  		printf("\n");
	  		return 1;
  }
  return 1;
}	

void parse_option(void)
{
  FILE *f;
  char linebuf[100];
  char *w, *s;
  int option;
  
  int csman_fd;
  struct in_addr waddr, laddr, lmask;
 
  if ((f=fopen(DEFAULT_CONF_FILE_PATH,"r")) == NULL)
    printf("can't open config file DEFAULT_CONF_FILE_PATH\n");
  
  while(fgets(linebuf,sizeof(linebuf),f) != NULL) {
    s = linebuf;
    w = (char*)next_word(&s);
    option = wordToOption(w);
    switch(option) {
			case EMPTY:
	    	continue;
				//break;
			case IGMPVERSION:
				printf("version:");
				if (EQUAL((w=(char *)next_word(&s))," ")) {
					version= DEFAULT_VERSION;
					printf("version set to default :%d\n",DEFAULT_VERSION);
				}
				else if(sscanf(w,"%d",&version) == 1) {
					printf("%d\n",version);
				}
				break;
			case IS_QUERIER:
				printf("isquerier:");
				if (EQUAL((w=(char *)next_word(&s))," ")) {
					querier=DEFAULT_ISQUERIER;
					printf("is_querier set to default :%d\n",DEFAULT_ISQUERIER);
				}
				else if(sscanf(w,"%d",&querier) == 1) {
					printf("%d\n",querier);
				}
				break;
		  case UPSTREAM:
			  w=(char *)next_word(&s);
				upstream = inet_addr(w);
				break;
						
			default:
				printf("unknown option\n");
		}
  }

  /* set upstream & downstream interface */
  csman_fd = open_csman(NULL, 0);
#ifndef _ENABLE_MULTI_WAN
  read_csman(csman_fd, CSID_S_LOCAL_WAN_IP, &waddr, sizeof(struct in_addr), CSM_R_ZERO);
#else
 /* For Multiwan , we use index to select the upstream interface exclusive */
  int multiWanIndex = 0;
  int mutualMode = 0;
  read_csman(csman_fd, CSID_C_LOCAL_MULTIWAN_IGMP_INTERFACE,&multiWanIndex,sizeof(int),CSM_R_ZERO);
  read_csman(csman_fd,CSID_C_LOCAL_MULTIWAN_MODE ,&mutualMode,sizeof(int),CSM_R_ZERO);
  if(0 == mutualMode)
  {
	  if(99 == multiWanIndex)
	  {
		  read_csman(csman_fd, CSID_S_LOCAL_WAN_IP, &waddr, sizeof(struct in_addr), CSM_R_ZERO);
	  }
	  else
	  {
		  read_csman(csman_fd,CSID_S_LOCAL_MULTIWAN_IP + multiWanIndex,&waddr,sizeof(struct in_addr), CSM_R_ZERO);
	  }
  }
  else
  {

	  if(99 == multiWanIndex || 13 == multiWanIndex || 0 == multiWanIndex)
	  {
		  read_csman(csman_fd, CSID_S_LOCAL_WAN_IP, &waddr, sizeof(struct in_addr), CSM_R_ZERO);
	  }
	  else
	  {
		  read_csman(csman_fd,CSID_S_LOCAL_MULTIWAN_IP + multiWanIndex,&waddr,sizeof(struct in_addr), CSM_R_ZERO);
	  }

  }
#endif
  read_csman(csman_fd, CSID_C_LOCAL_LANIP,  &laddr, sizeof(struct in_addr), CSM_R_ZERO);
  read_csman(csman_fd, CSID_C_LOCAL_LANNM,  &lmask, sizeof(struct in_addr), CSM_R_ZERO);
  close(csman_fd);

  upstream   = waddr.s_addr;
  downstream = laddr.s_addr;
  mask       = lmask.s_addr;
  subnet     = downstream & mask;

  //printf("wanip:%s\n", inet_ntoa(upstream));
  //printf("lanip:%s\n", inet_ntoa(downstream));
}

int
main(int argc, char *argv[])
{
	struct sockaddr_in *psin;
	struct ip_mreq mreq;
	interface_list_t *ifl, *ifp;
	char cmd[BUF_CMD];
	vifi_t vifi;
	log_level = LOG_DEBUG;
	
	/* Initialize */
	signal(SIGUSR1, done);
	signal(SIGKILL, done);
	signal(SIGABRT, done);
	signal(SIGTERM, done);
	signal(SIGHUP,	done);
	//write_pid();
	
	/*parse option on config file*/
	parse_option();
	
	/* Create and initialize the router */
	igmprt_init(&router);
	k_init_proxy(((igmp_router_t*)&router)->igmprt_socket);

	/* Add membership to ALL_ROUTERS and ALL_ROUTERS_V3 on this interface */
	mreq.imr_multiaddr.s_addr = inet_addr(IGMP_ALL_ROUTERS);
	mreq.imr_interface.s_addr = downstream;/*htonl(0);*/
	setsockopt(((igmp_router_t*)&router)->igmprt_socket, IPPROTO_IP, IP_ADD_MEMBERSHIP, (void*)&mreq, sizeof(mreq));
	
	mreq.imr_multiaddr.s_addr = inet_addr(IGMP_ALL_ROUTERS_V3);
	mreq.imr_interface.s_addr = downstream;/*htonl(0);*/
	setsockopt(((igmp_router_t*)&router)->igmprt_socket, IPPROTO_IP, IP_ADD_MEMBERSHIP, (void*)&mreq, sizeof(mreq));
	
	/* Add all the multicast enabled ipv4 interfaces */
	numvifs = 0;
	ifl = get_interface_list(AF_INET, IFF_MULTICAST, IFF_LOOPBACK);
	for (vifi=0, ifp=ifl; ifp; ifp=ifp->ifl_next, ++vifi)	{
		psin = (struct sockaddr_in*)&ifp->ifl_addr;
		if (psin->sin_addr.s_addr == upstream) {
			continue;
		}
		igmprt_interface_add(&router, psin->sin_addr, ifp->ifl_name, vifi);
		k_proxy_add_vif(((igmp_router_t*)&router)->igmprt_socket, psin->sin_addr.s_addr, vifi);
		++numvifs;
	}
	free_interface_list(ifl);

	/* Print the status of the router */
	igmprt_print(&router);
	
	/* Start the router */
	igmprt_start(&router);

  while (go_on) {
#if 0  	
		/* Read and process commands */
		printf("> ");
		fflush(stdout);
		if (fgets(cmd, BUF_CMD, stdin) != NULL) {
			go_on = process_cmd(cmd);
		}
#endif		
		sleep(1);
	}

	/* Done */
	printf("Router IGMP prepare to terminal.\n");
	done(0);
	return 0;
}

void udp_traffic_thread(void *arg)
{
	igmp_router_t *igmprt = (igmp_router_t*)arg;
	igmp_group_t *mcgroup;
	igmp_src_t *mcsrc;
	igmp_interface_t *ifp_lan;
	struct ip *iphdr;
	struct udphdr *udphdr;
	struct sockaddr_in sadr, dadr;
	int rcvlen, iphlen, on = 1, csman_fd;
	int udp_mfc_socket;
	socklen_t dummy = 0;
	unsigned long ssdpaddr = inet_addr("239.255.255.250");
	

	if (igmprt->igmprt_flag_udpproxy == 0)
		return;

	memset((char*)&sadr, 0, sizeof(sadr));
  sadr.sin_family      = AF_INET;
  sadr.sin_port        = 0;
  sadr.sin_addr.s_addr = downstream;

  ifp_lan = igmprt_interface_lookup(igmprt, sadr.sin_addr);

	if ((udp_mfc_socket = socket(AF_INET, SOCK_RAW, IPPROTO_UDP)) < 0)
		perror("socket() udp_mfc_socket");

	if (bind(udp_mfc_socket, (struct sockaddr*)&sadr, sizeof(sadr)) < 0)
		perror("bind() udp_mfc_socket");

	if (setsockopt(udp_mfc_socket, IPPROTO_IP, IP_HDRINCL, (void*)&on, sizeof(on)) < 0)
		perror("setsockopt() IP_HDRINCL");

	for (;;) {
		memset((char*)rcv_buf, 0, RCV_BUF_SIZE);
		rcvlen = recvfrom(igmprt->igmprt_up_socket, rcv_buf, RCV_BUF_SIZE, 0, NULL, &dummy);
		
		if (rcvlen <= sizeof(struct ip))
			continue;
		
		iphdr  = (struct iphdr*)rcv_buf;
		iphlen = iphdr->ip_hl << 2;		
		udphdr = (struct udphdr*)(iphdr + iphlen);
		
		// Set filter conditions -----------------------
		if (iphdr->ip_dst.s_addr == ssdpaddr)
			continue;
				
		if (!IN_MULTICAST(ntohl(iphdr->ip_dst.s_addr)))
			continue;
		
		if (!VALID_ADDR(iphdr->ip_src))
			continue;
			
		if ((mcgroup = igmp_interface_group_lookup(ifp_lan, iphdr->ip_dst)) == NULL)
			continue;

		if (mcgroup->igmpg_fmode == IGMP_FMODE_INCLUDE) {
			if ((mcsrc = igmp_group_src_lookup(mcgroup, iphdr->ip_src)) == NULL) {
				continue;
			}
			if (mcsrc->igmps_fstate == FALSE) {
				continue;
			}
		}				
				
		if (mcgroup->igmpg_fmode == IGMP_FMODE_EXCLUDE) {
			if ((mcsrc = igmp_group_src_lookup(mcgroup, iphdr->ip_src)) != NULL) {
				if (mcsrc->igmps_fstate == FALSE)
					continue;
			}
		}
		// -------------------------------------------------
		
		memset((char*)&dadr, 0, sizeof(dadr));
		dadr.sin_family = AF_INET;
		dadr.sin_port   = udphdr->uh_dport;    
		dadr.sin_addr 	= iphdr->ip_dst;    
		sendto(udp_mfc_socket, rcv_buf, rcvlen, 0, (struct sockaddr*)&dadr, sizeof(dadr));
	}
	
	close_csman(csman_fd);
	close(udp_mfc_socket);
}

igmp_report_t*
igmp_parse_pack(void *pack)
{
	igmp_report_t *report = (igmp_report_t*)report_buf;
	int gi, si, ngp, nsrc;
	size_t mv;
	
	memset(report_buf, 0, sizeof(*report_buf)); mv=0;
	memcpy(&report->igmpr_type,    pack,    1); ++mv;  // u_char, 1 byte
	memcpy(&report->igmpr_code,    pack+mv, 1); ++mv;  // u_char, 1 byte
	memcpy(&report->igmpr_cksum,   pack+mv, 2); mv+=2; // u_short, 2 byte
	memcpy(&report->igmpr_rsv,     pack+mv, 2); mv+=2; // u_short, 2 byte
	memcpy(&report->igmpr_numgrps, pack+mv, 2); mv+=2; // u_short, 2 byte
	ngp = ntohs(report->igmpr_numgrps);
	for (gi=0; gi<ngp; ++gi) {
		memcpy(&report->igmpr_group[gi].igmpg_type,    pack+mv, 1); ++mv;  // u_char, 1 byte
		memcpy(&report->igmpr_group[gi].igmpg_datalen, pack+mv, 1); ++mv;  // u_char, 1 byte
		memcpy(&report->igmpr_group[gi].igmpg_numsrc,  pack+mv, 2); mv+=2; // u_short, 2 byte
		memcpy(&report->igmpr_group[gi].igmpg_group,   pack+mv, 4); mv+=4; // struct in_addr, 4 byte
		nsrc = ntohs(report->igmpr_group[gi].igmpg_numsrc);
		for (si=0; si<nsrc; ++si) {
			memcpy(&report->igmpr_group[gi].igmpg_sources[si], pack+mv, 4);	mv+=4; // struct in_addr, 4 byte
		}
	}	
	
	return report;
}
